using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.IO.Compression;
using System.Reflection;

namespace CodeArchitects
{
   /// <summary>
   /// Provides the service to uncompress library DLLs on-the-fly.
   /// </summary>
   public static class AssemblyUnzipper
   {
      // the special signature appended to compressed data
      const int SIGNATURE = 0x124589FD;
      // the name of the EXE file
      static string ExeFile = null;
      // the dictionary of all compressed assemblies in the resource
      static Dictionary<string, AssemblyInfo> AsmInfos = new Dictionary<string, AssemblyInfo>();

      /// <summary>
      /// Initialize the assembly unzipper
      /// </summary>
      public static void Initialize()
      {
         // open the main exe file
         ExeFile = Assembly.GetCallingAssembly().Location;

         using ( FileStream stream = new FileStream(ExeFile, FileMode.Open, FileAccess.Read, FileShare.ReadWrite) )
         {
            BinaryReader reader = new BinaryReader(stream);
            // check that this file contains compressed data, exit if not
            int endOfData = (int) stream.Length - 8;
            stream.Position = endOfData;
            int startOfData = reader.ReadInt32();
            if ( reader.ReadInt32() != SIGNATURE )
               return;
            // seek to the beginning of compressed data
            stream.Position = startOfData;

            // read data about individial assemblies
            while ( stream.Position < endOfData )
            {
               AssemblyInfo info = new AssemblyInfo();
               // read assembly name (which is preceded by its length)
               info.Name = reader.ReadString();
               info.RealSize = reader.ReadInt32();
               info.CompressedSize = reader.ReadInt32();
               info.Offset = (int) stream.Position;
               AsmInfos.Add(info.Name, info);
               // skip the compressed data
               stream.Position = info.Offset + info.CompressedSize;

               // Debug.WriteLine("Found compressed assembly: " + info.Name);
            }
         }

         // install handler for AssemblyResolve event
         AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(CurrentDomain_AssemblyResolve);
      }

      // the handler for AssemblyResolve event

      static Assembly CurrentDomain_AssemblyResolve(object sender, ResolveEventArgs e)
      {
         // find the assembly with given name, cause error if not found
         AssemblyInfo info = null;
         if ( AsmInfos.TryGetValue(e.Name, out info) )
            return ExtractAssembly(info);
         // signal error
         Debug.WriteLine("Failed to uncompress assembly " + info.Name);
         return null;
      }

      // extract and uncompress an assembly

      private static Assembly ExtractAssembly(AssemblyInfo info)
      {
         // extract the bytes
         using ( FileStream stream = new FileStream(ExeFile, FileMode.Open, FileAccess.Read, FileShare.ReadWrite) )
         {
            // read the compressed data
            byte[] bytes = new byte[info.CompressedSize];
            stream.Position = info.Offset;
            stream.Read(bytes, 0, bytes.Length);
            // uncompress it and return 
            byte[] buffer = UncompressBytes(bytes, info.RealSize);
            Assembly asm = Assembly.Load(buffer);
            return asm;
         }
      }
      
      // unzip a byte array

      private static byte[] UncompressBytes(byte[] bytes, int realSize)
      {
         byte[] buffer = new byte[realSize];

         MemoryStream ms = new MemoryStream(bytes);
         using ( GZipStream zipStream = new GZipStream(ms, CompressionMode.Decompress) )
         {
            // read the uncompressed resource
            int bytesRead = 0;
            while ( bytesRead < buffer.Length )
               bytesRead += zipStream.Read(buffer, bytesRead, buffer.Length - bytesRead);
         }
         return buffer;
      }

      // the AssemblyInfo private class

      private class AssemblyInfo
      {
         public string Name;
         public int RealSize;
         public int CompressedSize;
         public int Offset;
      }
   }

}
